unit main;

{*******************************************************************************
*                                                                              *
*                      Test PA Sine for Delphi                                 *
*                                                                              *
*******************************************************************************}


interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls;

type
  TForm1 = class(TForm)
    ButtonSinePlay: TButton;
    Memo1: TMemo;
    ButtonClose: TButton;
    procedure ButtonSinePlayClick(Sender: TObject);
    procedure ButtonCloseClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    function testmain : Integer;
  end;

var
  Form1: TForm1;



implementation

uses
  PortAudio;

{$R *.dfm}
type
  paTestData = Record
    sine        : Array [0..199] of Single;
    left_phase  : Integer;
    right_phase : Integer;
  end;

const
  NUM_SECONDS        = 5;
  SAMPLE_RATE        = 44100;
  FRAMES_PER_BUFFER  = 64;

  M_PI  = (3.14159265);
  TABLE_SIZE  = (200);

var
    outputParameters:TPaStreamParameters;


{* This routine will be called by the PortAudio engine when audio is needed.
** It may called at interrupt level on some machines so don't do anything
** that could mess up the system like calling malloc() or free().
*}
function patestCallback( const inputBuffer : Pointer;
                         outputBuffer      : Pointer;
                         framesPerBuffer   : Cardinal;
                         const timeInfo    : PPaStreamCallbackTimeInfo;
                         statusFlags       : TPaStreamCallbackFlags;
                         userData          : Pointer ):Integer;  cdecl;
var
  data   : ^paTestData;
  outbuf : ^Single;
  i      : Cardinal;
begin
  data  := userData;
  outbuf:= outputBuffer;

  for i:= 0 to framesPerBuffer - 1 do
    begin
      outbuf^ := data^.sine[data^.left_phase];  //* left */
      Inc(outbuf);
      outbuf^ := data^.sine[data^.right_phase];  //* right */
      Inc( outbuf);
      Inc( data^.left_phase);
      if( data^.left_phase >= TABLE_SIZE )
        then data^.left_phase := data^.left_phase - TABLE_SIZE;
      //* higher pitch so we can distinguish left and right. */
      data^.right_phase := data^.right_phase + 3;
      if( data^.right_phase >= TABLE_SIZE )
        then data^.right_phase := data^.right_phase - TABLE_SIZE;
    end;

  result:= paContinue;
end;


function TForm1.testmain : Integer;
label error;
var
  stream   : PPaStream;
  err      : TPaError;
  data     : paTestData;
  i, stime : Integer;
  outputParameters : TPaStreamParameters;
begin
  Memo1.Lines.Add( 'PortAudio Test: output sine wave.');
                   // SR = %d, BufSize = %d\n", SAMPLE_RATE, FRAMES_PER_BUFFER');

  //* initialise sinusodial wavetable */
  for i:= 0 to TABLE_SIZE - 1
    do data.sine[i]:= sin( i*1.0/TABLE_SIZE * M_PI * 2. );

  data.left_phase := 0;
  data.right_phase:= 0;

  err:= Pa_Initialize();
  if( err <> paNoError )then goto error;

  outputParameters.device:= Pa_GetDefaultOutputDevice(); //* default output device */
  outputParameters.channelCount:= 2;         //* stereo output */
  outputParameters.sampleFormat:= paFloat32; //* 32 bit floating point output */
  outputParameters.suggestedLatency:= Pa_GetDeviceInfo( outputParameters.device )^.defaultLowOutputLatency;
  outputParameters.hostApiSpecificStreamInfo:= NIL;

  err := Pa_OpenStream( stream,
                        PPaStreamParameters(Nil), //* no input */
                        @outputParameters,
                        SAMPLE_RATE,
                        FRAMES_PER_BUFFER,
                        paClipOff,      //* we won't output out of range samples so don't bother clipping them */
                        @patestCallback,
                        @data );
  if err <> paNoError then goto error;

  err:= Pa_StartStream( stream );
  if err <> paNoError then goto error;

  Memo1.Lines.Add( 'Play for 5 seconds.');
//  Pa_Sleep( NUM_SECONDS * 1000 );
  sTime:= GetTickCount();
  while (GetTickCount() - sTime) < 5000 do Application.ProcessMessages;


  err:= Pa_StopStream( stream);
  if( err <> paNoError )then goto error;

  err:= Pa_CloseStream( stream);
  if err <> paNoError then goto error;

  Pa_Terminate();
  Memo1.Lines.Add( 'Test finished.');

  Result := err;
  exit;

error:
  Pa_Terminate();
  Memo1.Lines.Add( 'An error occured while using the portaudio stream');
  Memo1.Lines.Add( 'Error number:  ' +IntToStr( err) );
//  fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );
  result:= err;
end;


procedure TForm1.ButtonSinePlayClick(Sender: TObject);
begin
  ButtonSinePlay.Enabled:= FALSE;
  ButtonClose.Enabled:= FALSE;

  testmain();

  ButtonSinePlay.Enabled:= TRUE;
  ButtonClose.Enabled:= TRUE;
end;

procedure TForm1.ButtonCloseClick(Sender: TObject);
begin
  close;
end;


end.
